@@ -7,3 +7,7 @@ from arkouda.pdarraycreation import *
 from arkouda.numeric import *
 from arkouda.pdarrayIO import *
 from arkouda.groupbyclass import *
+from arkouda.strings import *
+from arkouda.join import *
+from arkouda.registration import *
+from arkouda.categorical import *
