@@ -13,7 +13,7 @@ def parse_single_value(msg):
     Attempt to convert a scalar return value from the arkouda server to a numpy
     scalar in Python. The user should not call this function directly.
     """
-    dtname, value = msg.split()
+    dtname, value = msg.split(maxsplit=1)
     mydtype = dtype(dtname)
     if mydtype == bool:
         if value == "True":
@@ -23,18 +23,20 @@ def parse_single_value(msg):
         else:
             raise ValueError("unsupported value from server {} {}".format(mydtype.name, value))
     try:
+        if mydtype == str_:
+            return mydtype.type(value.strip('"'))
         return mydtype.type(value)
     except:
         raise ValueError("unsupported value from server {} {}".format(mydtype.name, value))
-    
+
 # class for the pdarray
 class pdarray:
     """
-    The basic arkouda array class. This class contains only the 
-    attributies of the array; the data resides on the arkouda 
-    server. When a server operation results in a new array, arkouda 
-    will create a pdarray instance that points to the array data on 
-    the server. As such, the user should not initialize pdarray 
+    The basic arkouda array class. This class contains only the
+    attributies of the array; the data resides on the arkouda
+    server. When a server operation results in a new array, arkouda
+    will create a pdarray instance that points to the array data on
+    the server. As such, the user should not initialize pdarray
     instances directly.
 
     Attributes
@@ -55,7 +57,10 @@ class pdarray:
 
     BinOps = frozenset(["+", "-", "*", "/", "//", "%", "<", ">", "<=", ">=", "!=", "==", "&", "|", "^", "<<", ">>","**"])
     OpEqOps = frozenset(["+=", "-=", "*=", "/=", "//=", "&=", "|=", "^=", "<<=", ">>=","**="])
-    
+    objtype = "pdarray"
+
+    __array_priority__ = 1000
+
     def __init__(self, name, mydtype, size, ndim, shape, itemsize):
         self.name = name
         self.dtype = dtype(mydtype)
@@ -70,6 +75,11 @@ class pdarray:
         except:
             pass
 
+    def __bool__(self):
+        if self.size != 1:
+            raise ValueError("The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()")
+        return bool(self[0])
+
     def __len__(self):
         return self.shape[0]
 
@@ -80,10 +90,10 @@ class pdarray:
     def __repr__(self):
         global pdarrayIterTresh
         return generic_msg("repr {} {}".format(self.name,pdarrayIterThresh))
- 
+
     def format_other(self, other):
         """
-        Attempt to cast scalar other to the element dtype of this pdarray, 
+        Attempt to cast scalar other to the element dtype of this pdarray,
         and print the resulting value to a string (e.g. for sending to a
         server command). The user should not call this function directly.
         """
@@ -95,7 +105,7 @@ class pdarray:
             return str(other)
         fmt = NUMBER_FORMAT_STRINGS[self.dtype.name]
         return fmt.format(other)
-        
+
     # binary operators
     def binop(self, other, op):
         if op not in self.BinOps:
@@ -204,10 +214,10 @@ class pdarray:
     def __rxor__(self, other):
         return self.r_binop(other, "^")
 
-    def __pow__(self,other): 
+    def __pow__(self,other):
         return self.binop(other,"**")
-    
-    def __rpow__(self,other): 
+
+    def __rpow__(self,other):
         return self.r_binop(other,"**")
 
     # overloaded comparison operators
@@ -257,7 +267,7 @@ class pdarray:
             other = self.dtype.type(other)
         except: # Can't cast other as dtype of pdarray
             raise TypeError("Unhandled scalar type: {} ({})".format(other, type(other)))
-            
+
         msg = "opeqvs {} {} {} {}".format(op, self.name, self.dtype.name, self.format_other(other))
         generic_msg(msg)
         return self
@@ -307,13 +317,17 @@ class pdarray:
     # overload a[] to treat like list
     def __getitem__(self, key):
         if np.isscalar(key) and resolve_scalar_dtype(key) == 'int64':
+            orig_key = key
+            if key < 0:
+                # Interpret negative key as offset from end of array
+                key += self.size
             if (key >= 0 and key < self.size):
                 repMsg = generic_msg("[int] {} {}".format(self.name, key))
                 fields = repMsg.split()
                 # value = fields[2]
                 return parse_single_value(' '.join(fields[1:]))
             else:
-                raise IndexError("[int] {} is out of bounds with size {}".format(key,self.size))
+                raise IndexError("[int] {} is out of bounds with size {}".format(orig_key,self.size))
         if isinstance(key, slice):
             (start,stop,stride) = key.indices(self.size)
             if verbose: print(start,stop,stride)
@@ -331,11 +345,15 @@ class pdarray:
             raise TypeError("Unhandled key type: {} ({})".format(key, type(key)))
 
     def __setitem__(self, key, value):
-        if isinstance(key, int):
+        if np.isscalar(key) and resolve_scalar_dtype(key) == 'int64':
+            orig_key = key
+            if key < 0:
+                # Interpret negative key as offset from end of array
+                key += self.size
             if (key >= 0 and key < self.size):
                 generic_msg("[int]=val {} {} {} {}".format(self.name, key, self.dtype.name, self.format_other(value)))
             else:
-                raise IndexError("index {} is out of bounds with size {}".format(key,self.size))
+                raise IndexError("index {} is out of bounds with size {}".format(orig_key,self.size))
         elif isinstance(key, pdarray):
             if isinstance(value, pdarray):
                 generic_msg("[pdarray]=pdarray {} {} {}".format(self.name,key.name,value.name))
@@ -351,12 +369,6 @@ class pdarray:
         else:
             raise TypeError("Unhandled key type: {} ({})".format(key, type(key)))
 
-    def __iter__(self):
-        # to_ndarray will error if array is too large to bring back
-        a = self.to_ndarray()
-        for x in a:
-            yield x
-            
     def fill(self, value):
         """
         Fill the array (in place) with a constant value.
@@ -368,7 +380,7 @@ class pdarray:
         Return True iff any element of the array evaluates to True.
         """
         return any(self)
-    
+
     def all(self):
         """
         Return True iff all elements of the array evaluate to True.
@@ -380,56 +392,56 @@ class pdarray:
         Return True iff the array is monotonically non-decreasing.
         """
         return is_sorted(self)
-    
+
     def sum(self):
         """
         Return the sum of all elements in the array.
         """
         return sum(self)
-    
+
     def prod(self):
         """
         Return the product of all elements in the array. Return value is
         always a float.
         """
         return prod(self)
-    
+
     def min(self):
         """
         Return the minimum value of the array.
         """
         return min(self)
-    
+
     def max(self):
         """
         Return the maximum value of the array.
         """
         return max(self)
-    
+
     def argmin(self):
         """
         Return the index of the first minimum value of the array.
         """
         return argmin(self)
-    
+
     def argmax(self):
         """
         Return the index of the first maximum value of the array.
         """
         return argmax(self)
-    
+
     def mean(self):
         """
         Return the mean of the array.
         """
         return mean(self)
-    
+
     def var(self, ddof=0):
         """
         Compute the variance. See ``arkouda.var`` for details.
         """
         return var(self, ddof=ddof)
-    
+
     def std(self, ddof=0):
         """
         Compute the standard deviation. See ``arkouda.std`` for details.
@@ -439,7 +451,7 @@ class pdarray:
     def to_ndarray(self):
         """
         Convert the array to a np.ndarray, transferring array data from the
-        arkouda server to Python. If the array exceeds a builtin size limit, 
+        arkouda server to Python. If the array exceeds a builtin size limit,
         a RuntimeError is raised.
 
         Returns
@@ -485,6 +497,64 @@ class pdarray:
         # Return a numpy ndarray
         return np.array(struct.unpack(fmt, rep_msg))
 
+    def to_cuda(self):
+        """
+        Convert the array to a Numba DeviceND array, transferring array data from the
+        arkouda server to Python via ndarray. If the array exceeds a builtin size limit,
+        a RuntimeError is raised.
+
+        Returns
+        -------
+        numba.DeviceNDArray
+            A Numba ndarray with the same attributes and data as the pdarray; on GPU
+
+        Notes
+        -----
+        The number of bytes in the array cannot exceed ``arkouda.maxTransferBytes``,
+        otherwise a ``RuntimeError`` will be raised. This is to protect the user
+        from overflowing the memory of the system on which the Python client
+        is running, under the assumption that the server is running on a
+        distributed system with much more memory than the client. The user
+        may override this limit by setting ak.maxTransferBytes to a larger
+        value, but proceed with caution.
+
+        See Also
+        --------
+        array
+
+        Examples
+        --------
+        >>> a = ak.arange(0, 5, 1)
+        >>> a.to_cuda()
+        array([0, 1, 2, 3, 4])
+
+        >>> type(a.to_cuda())
+        numpy.devicendarray
+        """
+        try:
+            from numba import cuda
+            if not(cuda.is_available()):
+                raise ImportError('CUDA is not available. Check for the CUDA toolkit and ensure a GPU is installed.')
+                return
+        except:
+            raise ModuleNotFoundError('Numba is not enabled or installed and is required for GPU support.')
+            return
+
+        # Total number of bytes in the array data
+        arraybytes = self.size * self.dtype.itemsize
+        # Guard against overflowing client memory
+        if arraybytes > maxTransferBytes:
+            raise RuntimeError("Array exceeds allowed size for transfer. Increase ak.maxTransferBytes to allow")
+        # The reply from the server will be a bytes object
+        rep_msg = generic_msg("tondarray {}".format(self.name), recv_bytes=True)
+        # Make sure the received data has the expected length
+        if len(rep_msg) != self.size*self.dtype.itemsize:
+            raise RuntimeError("Expected {} bytes but received {}".format(self.size*self.dtype.itemsize, len(rep_msg)))
+        # Use struct to interpret bytes as a big-endian numeric array
+        fmt = '>{:n}{}'.format(self.size, structDtypeCodes[self.dtype.name])
+        # Return a numba devicendarray
+        return cuda.to_device(struct.unpack(fmt, rep_msg))
+
     def save(self, prefix_path, dataset='array', mode='truncate'):
         """
         Save the pdarray to HDF5. The result is a collection of HDF5 files,
@@ -585,7 +655,7 @@ def all(pda):
         return parse_single_value(repMsg)
     else:
         raise TypeError("must be pdarray {}".format(pda))
-    
+
 def is_sorted(pda):
     """
     Return True iff the array is monotonically non-decreasing.
@@ -636,7 +706,7 @@ def max(pda):
         return parse_single_value(repMsg)
     else:
         raise TypeError("must be pdarray {}".format(pda))
-    
+
 def argmin(pda):
     """
     Return the index of the first minimum value of the array.
